-- DOWN PROCEDURES
DROP PROCEDURE IF EXISTS p_class_unenroll_member
DROP PROCEDURE IF EXISTS p_class_enroll_member
DROP PROCEDURE IF EXISTS p_class_update
DROP PROCEDURE IF EXISTS p_class_create
DROP PROCEDURE IF EXISTS p_member_update
DROP PROCEDURE IF EXISTS p_member_create
DROP PROCEDURE IF EXISTS p_password_change
DROP PROCEDURE IF EXISTS p_password_validate
GO

-- DOWN TRIGGERS
DROP TRIGGER IF EXISTS t_invoice_enrolled_member
DROP TRIGGER IF EXISTS t_max_class_size
DROP TRIGGER IF EXISTS t_encryptpassword
GO

-- DOWN FUNCTIONS
DROP FUNCTION IF EXISTS f_validate_password
DROP FUNCTION IF EXISTS f_generate_password
DROP FUNCTION IF EXISTS f_lcd_by_class_id
DROP FUNCTION IF EXISTS f_gcd_by_class_id
DROP FUNCTION IF EXISTS f_wordcloud_by_class_id
DROP FUNCTIOn IF EXISTS f_revenue_by_timeframe
DROP FUNCTION IF EXISTS f_revenue_by_leader_timeframe
DROP FUNCTION IF EXISTS f_revenue_by_leader_id
DROP FUNCTION IF EXISTS f_revenue_by_class_id
DROP FUNCTION IF EXISTS f_invoices_by_member_email
DROP FUNCTION IF EXISTS f_invoices_by_member_id
DROP FUNCTION IF EXISTS f_invoices_by_class_id
DROP FUNCTION IF EXISTS f_class_attendance_by_member_email
DROP FUNCTION IF EXISTS f_class_attendance_by_member_id
DROP FUNCTION IF EXISTS f_class_attendance_by_leader_email
DROP FUNCTION IF EXISTS f_class_attendance_by_leader_id
DROP FUNCTION IF EXISTS f_class_attendance_by_class_id
DROP FUNCTION IF EXISTS f_classes_by_leader_email
DROP FUNCTION IF EXISTS f_classes_by_leader_id
GO

-- DOWN VIEWS
DROP VIEW IF EXISTS v_invoices_all
DROP VIEW IF EXISTS v_class_enrollment
DROP VIEW IF EXISTS v_classes_happening_now
DROP VIEW IF EXISTS v_classes_prior
DROP VIEW IF EXISTS v_classes_upcoming
DROP VIEW IF EXISTS v_classes_all
DROP VIEW IF EXISTS v_members_demographics_only
DROP VIEW IF EXISTS v_member_list_contact_info
DROP VIEW IF EXISTS v_member_list_all_info
DROP VIEW IF EXISTS v_member_list
DROP VIEW IF EXISTS v_picklist_political_affiliations
DROP VIEW IF EXISTS v_picklist_pronouns
DROP VIEW IF EXISTS v_picklist_genders
DROP VIEW IF EXISTS v_picklist_races
DROP VIEW IF EXISTS v_picklist_countries
GO

-- DOWN CONSTRAINTS
IF EXISTS(SELECT * FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS where CONSTRAINT_NAME ='ck_invoice_over_zero')
    ALTER TABLE invoices DROP CONSTRAINT ck_invoice_over_zero

IF EXISTS(SELECT * FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS where CONSTRAINT_NAME ='fk_member_id')
    ALTER TABLE invoices DROP CONSTRAINT fk_member_id

IF EXISTS(SELECT * FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS where CONSTRAINT_NAME ='fk_class_id')
    ALTER TABLE invoices DROP CONSTRAINT fk_class_id

IF EXISTS(SELECT * FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS where CONSTRAINT_NAME ='fk_class_members_member_id')
    ALTER TABLE class_members DROP CONSTRAINT fk_class_members_member_id

IF EXISTS(SELECT * FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS where CONSTRAINT_NAME ='fk_class_members_class_id')
    ALTER TABLE class_members DROP CONSTRAINT fk_class_members_class_id

IF EXISTS(SELECT * FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS where CONSTRAINT_NAME ='fk_classes_leader')
    ALTER TABLE classes DROP CONSTRAINT fk_classes_leader

IF EXISTS(SELECT * FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS where CONSTRAINT_NAME ='fk_members_political_affiliation')
    ALTER TABLE members DROP CONSTRAINT fk_members_political_affiliation

IF EXISTS(SELECT * FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS where CONSTRAINT_NAME ='fk_members_pronoun')
    ALTER TABLE members DROP CONSTRAINT fk_members_pronoun

IF EXISTS(SELECT * FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS where CONSTRAINT_NAME ='fk_members_gender')
    ALTER TABLE members DROP CONSTRAINT fk_members_gender

IF EXISTS(SELECT * FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS where CONSTRAINT_NAME ='fk_members_race')
    ALTER TABLE members DROP CONSTRAINT fk_members_race

IF EXISTS(SELECT * FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS where CONSTRAINT_NAME ='fk_members_origin_country')
    ALTER TABLE members DROP CONSTRAINT fk_members_origin_country
GO

-- DOWN TABLES
DROP TABLE IF EXISTS invoices
DROP TABLE IF EXISTS class_members
DROP TABLE IF EXISTS classes
DROP TABLE IF EXISTS members
DROP TABLE IF EXISTS political_affiliations
DROP TABLE IF EXISTS pronouns
DROP TABLE IF EXISTS genders
DROP TABLE IF EXISTS races
DROP TABLE IF EXISTS countries
GO

-- UP TABLES
CREATE TABLE countries
    ( country VARCHAR(255) PRIMARY KEY )
CREATE TABLE races
    ( race VARCHAR(255) PRIMARY KEY )
CREATE TABLE genders
    ( gender VARCHAR(255) PRIMARY KEY )
CREATE TABLE pronouns
    ( pronoun VARCHAR(255) PRIMARY KEY )
CREATE TABLE political_affiliations
    ( political_affiliation VARCHAR(255) PRIMARY KEY )

CREATE TABLE members
    (
        member_id INT IDENTITY PRIMARY KEY,
        member_email VARCHAR(255) UNIQUE NOT NULL,
        member_firstname VARCHAR(80) NOT NULL,
        member_lastname VARCHAR(80) NOT NULL,
        member_religion VARCHAR(80) NOT NULL,
        member_birthday DATE,
        member_phone VARCHAR(50),
        member_language_1 VARCHAR(100) NOT NULL,
        member_language_2 VARCHAR(100),
        member_language_3 VARCHAR(100),
        member_language_4 VARCHAR(100),
        member_language_5 VARCHAR(100),
        member_origin_country VARCHAR(255) NOT NULL,
        member_race VARCHAR(255) NOT NULL,
        member_gender VARCHAR(255) NOT NULL,
        member_pronoun VARCHAR(255) NOT NULL,
        member_political_affiliation VARCHAR(255) NOT NULL,
        member_password VARCHAR(2000) NOT NULL,
        member_salt VARCHAR(36) NOT NULL
    )

CREATE TABLE classes
    (
        class_id INT IDENTITY PRIMARY KEY,
        class_name VARCHAR(255) NOT NULL,
        class_leader_id INT NOT NULL,
        class_date_time_start DATETIME NOT NULL,
        class_date_time_end DATETIME NOT NULL,
        class_description VARCHAR(8000) NOT NULL,
        class_price DECIMAL(10,2) NOT NULL,
        class_max_size DECIMAL(4,0) NOT NULL
    ) 

CREATE TABLE class_members
    (
        class_id INT NOT NULL,
        member_id INT NOT NULL
    )

CREATE TABLE invoices
    (
        invoice_id INT IDENTITY PRIMARY KEY,
        invoice_price DECIMAL(10,2) NOT NULL,
        invoice_class_id INT NOT NULL,
        invoice_member_id INT NOT NULL
    )
GO

-- UP CONSTRAINTS
ALTER TABLE members ADD CONSTRAINT fk_members_origin_country
    FOREIGN KEY (member_origin_country) REFERENCES countries(country)

ALTER TABLE members ADD CONSTRAINT fk_members_race
    FOREIGN KEY (member_race) REFERENCES races(race)

ALTER TABLE members ADD CONSTRAINT fk_members_gender
    FOREIGN KEY (member_gender) REFERENCES genders(gender)

ALTER TABLE members ADD CONSTRAINT fk_members_pronoun
    FOREIGN KEY (member_pronoun) REFERENCES pronouns(pronoun)

ALTER TABLE members ADD CONSTRAINT fk_members_political_affiliation
    FOREIGN KEY (member_political_affiliation) REFERENCES political_affiliations(political_affiliation)

ALTER TABLE classes ADD CONSTRAINT fk_classes_leader
    FOREIGN KEY (class_leader_id) REFERENCES members(member_id)

ALTER TABLE class_members ADD CONSTRAINT fk_class_members_class_id
    FOREIGN KEY (class_id) REFERENCES classes(class_id)

ALTER TABLE class_members ADD CONSTRAINT fk_class_members_member_id
    FOREIGN KEY (member_id) REFERENCES members(member_id)

ALTER TABLE class_members ADD CONSTRAINT pk_class_members_id
    PRIMARY KEY (class_id, member_id)

ALTER TABLE invoices ADD CONSTRAINT fk_class_id
    FOREIGN KEY (invoice_class_id) REFERENCES classes(class_id)

ALTER TABLE invoices ADD CONSTRAINT fk_member_id
    FOREIGN KEY (invoice_member_id) REFERENCES members(member_id)

ALTER TABLE invoices ADD CONSTRAINT ck_invoice_over_zero
    CHECK (invoice_price > 0)

GO
-- UP VIEWS
CREATE VIEW v_picklist_countries AS
    (SELECT country countries FROM countries)
    GO
CREATE VIEW v_picklist_races AS
    (SELECT race races FROM races)
GO
CREATE VIEW v_picklist_genders AS
    (SELECT gender genders FROM genders)
GO
CREATE VIEW v_picklist_pronouns AS
    (SELECT pronoun pronouns FROM pronouns)
GO
CREATE VIEW v_picklist_political_affiliations AS
    (SELECT political_affiliation political_affiliations FROM political_affiliations)
GO

-- The member list views should not include the password column.  Even though we are storing it in an encrypted format, there's no good reason to expose the encrypted phrase.
CREATE VIEW v_member_list_all_info AS
    (SELECT member_id, member_email, member_firstname, member_lastname, member_religion, member_birthday, member_phone, member_language_1, member_language_2, member_language_3, member_language_4, member_language_5, member_origin_country, member_race, member_gender, member_pronoun, member_political_affiliation FROM members)
GO

-- This view only has contact information for promotional campaigns.  It is intended to convey very little personal demographics information.
CREATE VIEW v_member_list_contact_info AS
    (SELECT member_firstname, member_lastname, member_email, member_phone, member_birthday, member_language_1 FROM members)
GO

-- The demographics only view is intended to remove personally identifiable information.  This allows us to perform or contract out data analytics without revealing the members.
-- It should be noted that for many people, you could still potentially tie this data back to a specific individual with a high level of confidence, so we must still be careful
-- in who has access to this data and who can see it.
CREATE VIEW v_members_demographics_only AS
    (SELECT member_id, member_religion, member_birthday, member_language_1, member_language_2, member_language_3, member_language_4, member_language_5,member_origin_country, member_race, member_gender, member_pronoun, member_political_affiliation FROM members)
GO

-- All classes, past and future
CREATE VIEW v_classes_all AS
    (SELECT * FROM classes)
GO

-- Classes scheduled
CREATE VIEW v_classes_upcoming AS
    (SELECT *, DATEDIFF(d,SYSUTCDATETIME(),class_date_time_start) AS Days_Until FROM classes WHERE class_date_time_start > SYSUTCDATETIME())
GO

-- Classes previously held
CREATE VIEW v_classes_prior AS
    (SELECT *, DATEDIFF(d,class_date_time_start,SYSUTCDATETIME()) as Days_Ago FROM classes WHERE class_date_time_end < SYSUTCDATETIME())
GO

-- Classes currently in session
CREATE VIEW v_classes_happening_now AS
    (SELECT * FROM classes WHERE class_date_time_start < SYSUTCDATETIME() AND class_date_time_end > SYSUTCDATETIME())
GO

-- Class enrollment report for all classes
CREATE VIEW v_class_enrollment AS
    (
        SELECT 
            c.class_name, c.class_description, m.member_firstname + ' ' + m.member_lastname AS member_name, m.member_email, m.member_phone, c.class_price,
            c.class_max_size, lead.member_firstname + ' ' + lead.member_lastname AS leader_name, c.class_date_time_start, c.class_date_time_end, c.class_id, m.member_id, c.class_leader_id               
        FROM class_members cm
        JOIN classes c ON cm.class_id = c.class_id
        JOIN members m ON cm.member_id = m.member_id
        JOIN members lead ON c.class_leader_id = lead.member_id
        )
GO

-- All Invoices
CREATE VIEW v_invoices_all AS 
    (
        SELECT i.invoice_id, i.invoice_price, c.class_name, c.class_description, m.member_firstname + ' ' + m.member_lastname AS member_name, m.member_email, m.member_phone,
         c.class_price, c.class_date_time_start, c.class_date_time_end, c.class_id, m.member_id, c.class_leader_id FROM invoices i
            JOIN classes c ON i.invoice_class_id = c.class_id
            JOIN members m ON i.invoice_member_id = m.member_id
    )
GO


-- UP FUNCTIONS
-- Let's see all the classes taught by a particular leader.  Returns a table of classes.  Select by id or email.
CREATE FUNCTION f_classes_by_leader_id
    (@leaderid AS INT) RETURNS TABLE AS 
    RETURN
    (SELECT * FROM classes WHERE class_leader_id = @leaderid)
GO

-- Let's see all the classes taught by a particular leader.  Returns a table of classes.  Select by id or email.
CREATE FUNCTION f_classes_by_leader_email
    (@leaderemail AS varchar(255)) RETURNS TABLE AS 
    RETURN
    (
        SELECT * FROM classes c 
        JOIN members m ON c.class_leader_id = m.member_id
        WHERE member_email = @leaderemail
    )
GO

-- Let's see all the members enrolled in a specific class
CREATE FUNCTION f_class_attendance_by_class_id
    (@classid AS INT) RETURNS TABLE AS
    RETURN
    (
        SELECT * FROM v_class_enrollment WHERE class_id = @classid
    )
GO

-- Let's see all the members enrolled for all classes conducted by a particular leader.  Select by id or email.
CREATE FUNCTION f_class_attendance_by_leader_id
    (@leaderid AS INT) RETURNS TABLE AS
    RETURN
    (
        SELECT * FROM v_class_enrollment WHERE class_leader_id = @leaderid
    )
GO

-- Let's see all the members enrolled for all classes conducted by a particular leader.  Select by id or email.
CREATE FUNCTION f_class_attendance_by_leader_email
    (@leaderemail AS varchar(255)) RETURNS TABLE AS 
    RETURN
    (
        SELECT c.class_name, c.class_description, m.member_firstname + ' ' + m.member_lastname AS member_name, m.member_email, m.member_phone, c.class_price,
            c.class_max_size, c.leader_name, c.class_date_time_start, c.class_date_time_end, c.class_id, m.member_id, c.class_leader_id
        FROM v_class_enrollment c 
        JOIN members m ON c.class_leader_id = m.member_id
        WHERE m.member_email = @leaderemail
    )
GO

-- All classes a member enrolled in.  Select by id or email.
CREATE FUNCTION f_class_attendance_by_member_id
    (@memberid AS INT) RETURNS TABLE AS
    RETURN
    (
        SELECT * FROM v_class_enrollment WHERE member_id = @memberid
    )
GO

-- All classes a member enrolled in.  Select by id or email.
CREATE FUNCTION f_class_attendance_by_member_email
    (@memberemail AS varchar(255)) RETURNS TABLE AS 
    RETURN
    (
        SELECT c.class_name, c.class_description, m.member_firstname + ' ' + m.member_lastname AS member_name, m.member_email, m.member_phone, c.class_price,
            c.class_max_size, c.leader_name, c.class_date_time_start, c.class_date_time_end, c.class_id, m.member_id, c.class_leader_id
        FROM v_class_enrollment c 
        JOIN members m ON c.member_id = m.member_id
        WHERE m.member_email = @memberemail
    )
GO

-- All invoices related to one class
CREATE FUNCTION f_invoices_by_class_id
    (@classid AS INT) RETURNS TABLE AS
    RETURN
    (
        SELECT * FROM v_invoices_all WHERE class_id = @classid
    )
GO

-- All invoices for a particular member.  Select by id or email.
CREATE FUNCTION f_invoices_by_member_id
    (@memberid AS INT) RETURNS TABLE AS
    RETURN
    (
        SELECT * FROM v_invoices_all WHERE member_id = @memberid
    )
GO

CREATE FUNCTION f_invoices_by_member_email
    (@memberemail AS varchar(255)) RETURNS TABLE AS 
    RETURN
    (
        SELECT i.invoice_id, i.invoice_price, i.class_name, i.class_description, i.member_name, i.member_email, i.member_phone, i.class_price, i.class_date_time_start, i.class_date_time_end, i.class_id, i.member_id, i.class_leader_id FROM v_invoices_all i
        JOIN members m ON i.member_id = m.member_id
        WHERE m.member_email = @memberemail
    )
GO

-- All invoices associated to a particular class.  Determines how much revenue this class produced.
CREATE FUNCTION f_revenue_by_class_id
    (@classid AS INT) RETURNS TABLE AS
    RETURN
        (
            SELECT SUM(i.invoice_price) AS Revenue_Generated, i.invoice_class_id, c.class_name, c.class_description, m.member_firstname + ' ' + m.member_lastname AS leader_name FROM invoices i
            JOIN classes c ON i.invoice_class_id = c.class_id
            JOIN members m ON c.class_leader_id = m.member_id
            WHERE invoice_class_id = @classid GROUP BY i.invoice_class_id, c.class_name, c.class_description, m.member_firstname, m.member_lastname
        )
GO

-- All invoices across all classes lead by a particular leader.  Determines how much revenue this leader generated.
CREATE FUNCTION f_revenue_by_leader_id
    (@leaderid AS INT) RETURNS TABLE AS
    RETURN
        (
            SELECT SUM(i.invoice_price) AS Revenue_Generated, m.member_firstname + ' ' + m.member_lastname AS leader_name FROM invoices i
            JOIN classes c ON i.invoice_class_id = c.class_id
            JOIN members m ON c.class_leader_id = m.member_id
            WHERE c.class_leader_id = @leaderid GROUP BY m.member_firstname, m.member_lastname
        )
GO

-- Displays the revenue generated within the timeframe, grouped by leader id.
CREATE FUNCTION f_revenue_by_leader_timeframe
    (@start_time AS DATETIME, @end_time AS DATETIME) RETURNS TABLE AS
    RETURN
        (
            SELECT SUM(i.invoice_price) AS Revenue_Generated, m.member_firstname + ' ' + m.member_lastname AS leader_name FROM invoices i
            JOIN classes c ON i.invoice_class_id = c.class_id
            JOIN members m ON c.class_leader_id = m.member_id
            WHERE c.class_date_time_start >= @start_time AND c.class_date_time_start <= @end_time
            GROUP BY m.member_firstname, m.member_lastname
        )
GO

-- Displays the total revenue generated within the timeframe
CREATE FUNCTION f_revenue_by_timeframe
    (@start_time AS DATETIME, @end_time AS DATETIME) RETURNS TABLE AS
    RETURN
        (
            SELECT SUM(i.invoice_price) AS Revenue_Generated FROM invoices i
            JOIN classes c ON i.invoice_class_id = c.class_id
            WHERE c.class_date_time_start >= @start_time AND c.class_date_time_start <= @end_time
        )
GO

-- Produces a word cloud compatible listing of all the demographics for a particular class
-- Unpivot requires all tables to be the same type.
-- Found a trick to simulate that here https://stackoverflow.com/questions/39066742/how-to-unpivot-a-column-with-different-data-type-sql-server
CREATE FUNCTION f_wordcloud_by_class_id
    (@classid AS INT) RETURNS TABLE AS
    RETURN
    (
        WITH demowide AS (SELECT cm.class_id, cm.member_id, 
            CONVERT(VARCHAR(255),m.member_religion) member_religion,
            CONVERT(VARCHAR(255),m.member_language_1) member_language_1,
            CONVERT(VARCHAR(255),m.member_language_2) member_language_2,
            CONVERT(VARCHAR(255),m.member_language_3) member_language_3,
            CONVERT(VARCHAR(255),m.member_language_4) member_language_4,
            CONVERT(VARCHAR(255),m.member_language_5) member_language_5,
            CONVERT(VARCHAR(255),m.member_origin_country) member_origin_country,
            CONVERT(VARCHAR(255),m.member_race) member_race,
            CONVERT(VARCHAR(255),m.member_gender) member_gender,
            CONVERT(VARCHAR(255),m.member_pronoun) member_pronoun,
            CONVERT(VARCHAR(255),m.member_political_affiliation) member_political_affiliation
            FROM class_members cm
        JOIN members m ON cm.member_id = m.member_id
        WHERE cm.class_id = @classid)
        SELECT demographic
        FROM (SELECT class_id, member_id, member_religion, member_language_1, member_language_2, member_language_3, member_language_4, member_language_5, 
        member_origin_country, member_race, member_gender, member_pronoun, member_political_affiliation FROM demowide) as src
        UNPIVOT
        (
            demographic FOR demograpics IN (member_religion, member_language_1, member_language_2, member_language_3, member_language_4, member_language_5,
            member_origin_country, member_race, member_gender, member_pronoun, member_political_affiliation)
        ) AS unpvt
        WHERE demographic != '' AND demographic IS NOT NULL
    )
GO

-- Displays the most common demographics
CREATE FUNCTION f_gcd_by_class_id
    (@classid AS INT, @topcount AS INT) RETURNS TABLE AS
    RETURN
        (select TOP (@topcount) demographic, count(*) as frequency from f_wordcloud_by_class_id (2) GROUP BY demographic ORDER BY frequency DESC)
GO

-- Displays the least common demographics
CREATE FUNCTION f_lcd_by_class_id
    (@classid AS INT, @bottomcount AS INT) RETURNS TABLE AS
    RETURN
        (select TOP (@bottomcount) demographic, count(*) as frequency from f_wordcloud_by_class_id (2) GROUP BY demographic ORDER BY frequency)
GO



-- Generate a hashed, salted password.  Based on https://www.mssqltips.com/sqlservertip/4037/storing-passwords-in-a-secure-way-in-a-sql-server-database/
CREATE FUNCTION f_generate_password
    (@password AS VARCHAR(255), @salt AS VARCHAR(36)) RETURNS VARCHAR(2000) AS
    BEGIN
        RETURN CAST(HASHBYTES('SHA2_512', @password + @salt) AS VARCHAR(2000))
    END
GO

-- Returns 1 if passwords match, returns 0 if passwords do not match.
-- Trick to return boolean value from https://stackoverflow.com/questions/157114/how-to-output-a-boolean-in-t-sql-based-on-the-content-of-a-column
CREATE FUNCTION f_validate_password
    (@password AS VARCHAR(255), @salt AS VARCHAR(36), @encryptedpassword AS VARCHAR(2000)) RETURNS INTEGER AS
    BEGIN
        DECLARE @incoming_encrypted VARCHAR(2000) = dbo.f_generate_password(@password, @salt)
        IF(@incoming_encrypted = @encryptedpassword)
            BEGIN
                RETURN CAST(1 AS BIT)
            END
        RETURN 0
    END
GO

-- UP BEFORE TRIGGERS
-- This trigger will check if a class is full before allowing a member to register.
-- First, it counts the current number of members registered for a class from the class_members table.
-- Then, it finds out the max class size from the classes table.
-- If the number of enrolled members is at or above capacity, disallow the insert.
-- We are not allowing API access (via stored procedures) for a member to transfer classes.
-- They will only be inserted or deleted, never updated.
CREATE TRIGGER t_max_class_size ON class_members
    INSTEAD OF INSERT AS
    BEGIN
    BEGIN TRANSACTION
        BEGIN TRY
            IF ((SELECT COUNT(*) FROM class_members WHERE class_id = (SELECT class_id FROM inserted)) >= (SELECT class_max_size FROM classes WHERE class_id = (SELECT class_id FROM inserted)))
            BEGIN
                ;
                THROW 50007, 'Insert error t_max_class_size: Class is full', 1
                ROLLBACK
            END
            ELSE BEGIN
            INSERT INTO class_members (class_id, member_id) (SELECT class_id, member_id FROM inserted);
                COMMIT
            END
        END TRY
        BEGIN CATCH
            ROLLBACK
            ;
            THROW
            END CATCH
    END
GO

-- This trigger ensures the member password is always encrypted with a unique salt
-- Supports putting in with a member_id present or not.  If member_id is present,
-- be sure to do a 'SET identity_insert dbo.members ON' first, otherwise it will fail.
CREATE TRIGGER t_encryptpassword ON members
    INSTEAD OF INSERT AS
    BEGIN
        DECLARE @newsalt UNIQUEIDENTIFIER = NEWID()
        DECLARE @encryptedpassword VARCHAR(2000) = dbo.f_generate_password((SELECT member_password FROM inserted), @newsalt)
        DECLARE @member_id_present INTEGER = (SELECT member_id FROM inserted)
        BEGIN
        IF (@member_id_present = 0)
            BEGIN
                INSERT INTO members (member_email, member_firstname, member_lastname, member_religion, member_birthday, member_phone, member_language_1, member_language_2, member_language_3, member_language_4, member_language_5, member_origin_country, member_race, member_gender, member_pronoun, member_political_affiliation, member_password, member_salt)
                    (SELECT member_email, member_firstname, member_lastname, member_religion, member_birthday, member_phone, member_language_1, member_language_2, member_language_3, member_language_4, member_language_5, member_origin_country, member_race, member_gender, member_pronoun, member_political_affiliation, @encryptedpassword, @newsalt FROM inserted)
            END ELSE
            BEGIN
                INSERT INTO members (member_id, member_email, member_firstname, member_lastname, member_religion, member_birthday, member_phone, member_language_1, member_language_2, member_language_3, member_language_4, member_language_5, member_origin_country, member_race, member_gender, member_pronoun, member_political_affiliation, member_password, member_salt)
                    (SELECT member_id, member_email, member_firstname, member_lastname, member_religion, member_birthday, member_phone, member_language_1, member_language_2, member_language_3, member_language_4, member_language_5, member_origin_country, member_race, member_gender, member_pronoun, member_political_affiliation, @encryptedpassword, @newsalt FROM inserted)            
            END

        END
    END
GO

-- UP AFTER TRIGGERS
-- This trigger is designed to automatically invoice a member upon registration.
-- It should only fire if the class is not free.  The business decided not to invoice free classes.
-- If that requirement changes in the future, we can simply remove the where clause.
CREATE TRIGGER t_invoice_enrolled_member ON class_members AFTER INSERT AS
BEGIN
    INSERT INTO invoices (invoice_class_id, invoice_member_id, invoice_price)
    SELECT i.class_id, i.member_id, c.class_price FROM inserted i
    JOIN classes c ON c.class_id = i.class_id
    WHERE c.class_price > 0
END
GO

-- UP STORED PROCEDURES
-- Verify if a password is correct.  We don't know the 'correct' password value, just if it matches after encrypting.
-- Must supply the guessed password and member_id value.
-- Returns 1 if passwords match, returns 0 if passwords do not match.
CREATE PROCEDURE p_password_validate
    (@password AS VARCHAR(255), @memberid AS INTEGER) AS
    BEGIN
        DECLARE @salt AS VARCHAR(36) = (SELECT member_salt FROM members WHERE member_id = @memberid)
        DECLARE @encryptedpassword AS VARCHAR(2000) = (SELECT member_password FROM members WHERE member_id = @memberid)
        DECLARE @pwcheck AS INTEGER
        EXEC @pwcheck = dbo.f_validate_password @password, @salt, @encryptedpassword
        RETURN @pwcheck
    END
GO

-- This procedure will change a member's password.  The correct old password must be supplied.
-- If the old password is incorrect, an exception will be thrown.
CREATE PROCEDURE p_password_change
    (@oldpassword AS VARCHAR(255), @newpassword AS VARCHAR(255), @memberid AS INTEGER) AS
    BEGIN
        DECLARE @pwcheck AS INTEGER
        EXEC @pwcheck = p_password_validate @oldpassword, @memberid
        IF (@pwcheck = 0)
            BEGIN
                THROW 50009, 'p_password_change: Incorrect old password', 1                
            END ELSE
            BEGIN
                DECLARE @newsalt UNIQUEIDENTIFIER = NEWID()
                DECLARE @encryptedpassword VARCHAR(2000) = dbo.f_generate_password(@newpassword, @newsalt)
                UPDATE members SET member_password = @encryptedpassword, member_salt = @newsalt WHERE member_id = @memberid
            END
    END
GO

-- Create a new member.  If member_id is present, be sure to do a 'SET identity_insert dbo.members ON' first, 
-- otherwise the insert will fail.  If you are allowing us to automatically determine the member_id (default behavior),
-- set @member_id to 0.
CREATE PROCEDURE p_member_create
        (@member_id INT,
        @member_email VARCHAR(255),
        @member_firstname VARCHAR(80),
        @member_lastname VARCHAR(80),
        @member_religion VARCHAR(80),
        @member_birthday DATE,
        @member_phone VARCHAR(50),
        @member_language_1 VARCHAR(100),
        @member_language_2 VARCHAR(100),
        @member_language_3 VARCHAR(100),
        @member_language_4 VARCHAR(100),
        @member_language_5 VARCHAR(100),
        @member_origin_country VARCHAR(255),
        @member_race VARCHAR(255),
        @member_gender VARCHAR(255),
        @member_pronoun VARCHAR(255),
        @member_political_affiliation VARCHAR(255),
        @member_password VARCHAR(2000))
        AS
        BEGIN
            IF (@member_id != 0)
            BEGIN
                INSERT INTO members (member_id,member_email,member_firstname, member_lastname, member_religion, member_birthday, member_phone, member_language_1, member_language_2, member_language_3, member_language_4, member_language_5, member_origin_country, member_race, member_gender, member_pronoun, member_political_affiliation, member_password)
                 VALUES (@member_id,@member_email,@member_firstname, @member_lastname, @member_religion, @member_birthday, @member_phone, @member_language_1, @member_language_2, @member_language_3, @member_language_4, @member_language_5, @member_origin_country, @member_race, @member_gender, @member_pronoun, @member_political_affiliation, @member_password)
            END ELSE
            BEGIN
                INSERT INTO members (member_email,member_firstname, member_lastname, member_religion, member_birthday, member_phone, member_language_1, member_language_2, member_language_3, member_language_4, member_language_5, member_origin_country, member_race, member_gender, member_pronoun, member_political_affiliation, member_password)
                 VALUES (@member_email,@member_firstname, @member_lastname, @member_religion, @member_birthday, @member_phone, @member_language_1, @member_language_2, @member_language_3, @member_language_4, @member_language_5, @member_origin_country, @member_race, @member_gender, @member_pronoun, @member_political_affiliation, @member_password)
            END
        END
GO

--Update a member’s information based on the passed in @member_id.  We do not do differential updates,
-- so you must supply all the fields.  Note that this procedure will not let you change the member’s password.
-- You must use p_password_change for that.
CREATE PROCEDURE p_member_update
        (@member_id INT,
        @member_email VARCHAR(255),
        @member_firstname VARCHAR(80),
        @member_lastname VARCHAR(80),
        @member_religion VARCHAR(80),
        @member_birthday DATE,
        @member_phone VARCHAR(50),
        @member_language_1 VARCHAR(100),
        @member_language_2 VARCHAR(100),
        @member_language_3 VARCHAR(100),
        @member_language_4 VARCHAR(100),
        @member_language_5 VARCHAR(100),
        @member_origin_country VARCHAR(255),
        @member_race VARCHAR(255),
        @member_gender VARCHAR(255),
        @member_pronoun VARCHAR(255),
        @member_political_affiliation VARCHAR(255))
        AS
        BEGIN
            UPDATE members SET
                member_email = @member_email,
                member_firstname = @member_firstname,
                member_lastname = @member_lastname,
                member_religion = @member_religion,
                member_birthday = @member_birthday,
                member_phone = @member_phone,
                member_language_1 = @member_language_1,
                member_language_2 = @member_language_2,
                member_language_3 = @member_language_3,
                member_language_4 = @member_language_4,
                member_language_5 = @member_language_5,
                member_origin_country = @member_origin_country,
                @member_race = @member_race,
                @member_gender = @member_gender,
                @member_pronoun = @member_pronoun,
                @member_political_affiliation = @member_political_affiliation
            WHERE member_id = @member_id
        END
GO

-- Create a new class.  If class_id is present, be sure to do a 'SET identity_insert dbo.classes ON' first, 
-- otherwise the insert will fail.  If you are allowing us to automatically determine the class_id (default behavior),
-- set @class_id to 0.
CREATE PROCEDURE p_class_create
        @class_id INT,
        @class_name VARCHAR(255),
        @class_leader_id INT,
        @class_date_time_start DATETIME,
        @class_date_time_end DATETIME,
        @class_description VARCHAR(8000),
        @class_price DECIMAL(10,2),
        @class_max_size DECIMAL(4,0)
        AS
        BEGIN
            IF (@class_id != 0)
            BEGIN
                INSERT INTO classes (class_id, class_name, class_leader_id, class_date_time_start, class_date_time_end, class_description, class_price, class_max_size)
                 VALUES (@class_id, @class_name, @class_leader_id, @class_date_time_start, @class_date_time_end, @class_description, @class_price, @class_max_size)
            END ELSE
            BEGIN
                INSERT INTO classes (class_name, class_leader_id, class_date_time_start, class_date_time_end, class_description, class_price, class_max_size)
                 VALUES (@class_name, @class_leader_id, @class_date_time_start, @class_date_time_end, @class_description, @class_price, @class_max_size)
            END
        END
GO

-- Update class information based on the passed in @class_id.  We do not do differential updates,
-- so you must supply all the fields.  Note that if you reduce the class_max_size, we do not unenroll any member.
-- That class will effectively be oversubscribed, but no additional members will be able to enroll.
CREATE PROCEDURE p_class_update
        @class_id INT,
        @class_name VARCHAR(255),
        @class_leader_id INT,
        @class_date_time_start DATETIME,
        @class_date_time_end DATETIME,
        @class_description VARCHAR(8000),
        @class_price DECIMAL(10,2),
        @class_max_size DECIMAL(4,0)
        AS
        BEGIN
            UPDATE classes SET
                class_name = @class_name,
                class_leader_id = @class_leader_id,
                class_date_time_start = @class_date_time_start,
                class_date_time_end = @class_date_time_end,
                class_description = @class_description,
                class_price = @class_price,
                class_max_size = @class_max_size
            WHERE class_id = @class_id
        END
GO

-- Enroll a member into a class
CREATE PROCEDURE p_class_enroll_member
        @class_id INT,
        @member_id INT
        AS
        BEGIN
            INSERT INTO class_members (class_id, member_id) VALUES (@class_id, @member_id)
        END
GO

-- Unenroll a member from a class.  Note that this procedure will not remove any associated invoices.
-- We need to wait for the business to determine how refunds/cancellations work before we will implement
-- that logic.  This delete action cannot be undone.
CREATE PROCEDURE p_class_unenroll_member
        @class_id INT,
        @member_id INT
        AS
        BEGIN
            DELETE FROM class_members WHERE class_id = @class_id AND member_id = @member_id
        END
GO
